package sgr

import (
	"fmt"
	"io"
	"os"
	"strings"
	"sync"

	"github.com/mattn/go-isatty"
)

var (
	// Formatter used when ANSI escape sequences are supported.
	monoFormatter      = Formatter{colorDepth: 1}
	defaultFormatter   = Formatter{colorDepth: 4}
	paletteFormatter   = Formatter{colorDepth: 8}
	truecolorFormatter = Formatter{colorDepth: 24}

	// Memoization for NewFormatter.
	newFormatterVal  *Formatter
	newFormatterOnce sync.Once
)

// A Formatter contains information about the capabilities of the output
// Writer, such as whether or not it supports ANSI escape sequences, and
// which sequences are available.
//
// Methods on this type all support being called with a nil pointer.  This
// state suppresses all escape sequences, and, in some cases, allows
// optimizations where strings are passed through directly.
type Formatter struct {
	colorDepth int
}

// NewFormatter returns a formatter suitable for use with Stdout.
func NewFormatter() *Formatter {
	newFormatterOnce.Do(func() {
		newFormatterVal = NewFormatterForFile(os.Stdout)
	})

	return newFormatterVal
}

// NewFormatterForFile returns a formatter suitable for use with
// the file.
func NewFormatterForFile(file *os.File) *Formatter {
	term := os.Getenv("TERM")
	if term == "dumb" || !isatty.IsTerminal(file.Fd()) {
		return nil
	}

	// https://no-color.org/
	if _, nocolor := os.LookupEnv("NO_COLOR"); nocolor {
		return &monoFormatter
	}

	// The environment variable COLORTERM overrides the color depth
	// inferred from TERM.  Check for a value indicating color support.
	if colorterm := os.Getenv("COLORTERM"); colorterm == "truecolor" || colorterm == "24bit" {
		return &truecolorFormatter
	} else if colorterm == "yes" || colorterm == "true" {
		return &paletteFormatter
	}

	return NewFormatterForTerm(term)
}

// NewFormatterForWriter returns a formatter suitable for use with
// the writer.
func NewFormatterForWriter(w io.Writer) *Formatter {
	file, ok := w.(*os.File)
	if !ok {
		return nil
	}

	return NewFormatterForFile(file)
}

// NewFormatterForTerm returns a formatter with the correct color depth based
// on the terminal identification string.
func NewFormatterForTerm(term string) *Formatter {
	if strings.HasSuffix(term, "-m") {
		return &monoFormatter
	}

	if strings.HasSuffix(term, "-256color") {
		return &paletteFormatter
	}

	if strings.HasSuffix(term, "-truecolor") || strings.HasSuffix(term, "-24bit") {
		return &truecolorFormatter
	}

	return &defaultFormatter
}

// NewFormatterWithANSI returns a formatter that will always use ANSI escape
// sequences.  This function exists for testing, and should not be used in
// normal code.
func NewFormatterWithANSI() *Formatter {
	return &defaultFormatter
}

// ColorDepth returns the bit-depth used by the terminal to represent colors.
func (f *Formatter) ColorDepth() int {
	if f == nil {
		return 0
	}

	return f.colorDepth
}

// Boldf returns an object which will show the formatted text in bold when
// printed using a function in package fmt.
func (f *Formatter) Boldf(format string, a ...interface{}) interface{} {
	if f == nil {
		return fmt.Sprintf(format, a...)
	}

	return &formatter{
		prefix: sgrBold,
		format: format,
		args:   a,
	}
}

// Dimf returns an object which will show the formatted text dimmed when
// printed using a function in package fmt.
func (f *Formatter) Dimf(format string, a ...interface{}) interface{} {
	if f == nil {
		return fmt.Sprintf(format, a...)
	}

	return &formatter{
		prefix: sgrDim,
		format: format,
		args:   a,
	}
}

// Redf returns an object which will show the formatted text in red when
// printed using a function in package fmt.
func (f *Formatter) Redf(format string, a ...interface{}) interface{} {
	if f == nil {
		return fmt.Sprintf(format, a...)
	}

	return &formatter{
		prefix: sgrRedFG,
		format: format,
		args:   a,
	}
}

// Greenf returns an object which will show the formatted text in green when
// printed using a function in package fmt.
func (f *Formatter) Greenf(format string, a ...interface{}) interface{} {
	if f == nil {
		return fmt.Sprintf(format, a...)
	}

	return &formatter{
		prefix: sgrGreenFG,
		format: format,
		args:   a,
	}
}

// Style returns an object which will show the formatted text in requested
// style when printed using a function in package fmt.
func (f *Formatter) Style(style Style, format string, a ...interface{}) interface{} {
	if f == nil {
		return fmt.Sprintf(format, a...)
	}

	return &formatter{
		prefix: []byte(style),
		format: format,
		args:   a,
	}
}

type formatter struct {
	prefix []byte
	format string
	args   []interface{}
}

func (s *formatter) Format(f fmt.State, v rune) {
	if len(s.args) == 0 {
		format(
			f,
			v,
			s.prefix,
			s.format,
		)

		return
	}

	// Check that the verb is supported.  Otherwise, print error message.
	if v != 's' && v != 'v' {
		// Match the error reporting from the fmt package
		fmt.Fprintf(f, "%%!%c(format string=%v)", v, s.format)

		return
	}

	f.Write(s.prefix)
	if _, ok := f.Width(); ok {
		_, _ = io.WriteString(f, "%!(BADWIDTH)")
	} else if _, ok = f.Precision(); ok {
		_, _ = io.WriteString(f, "%!(BADPREC)")
	}
	fmt.Fprintf(f, s.format, s.args...)
	f.Write(sgrReset)
}
