// Package palette provides mapping for 24-bit colors to the 8-bit ANSI SGR
// palette.
package palette

var (
	// Palette is the 256-color lookup table defined by ANSI for use with
	// SGR escape codes.
	Palette = [0x100]RGB{
		{0, 0, 0},
		{192, 0, 0},
		{0, 192, 0},
		{192, 192, 0},
		{0, 0, 192},
		{192, 0, 192},
		{0, 192, 192},
		{192, 192, 192},
		{128, 128, 128},
		{255, 0, 0},
		{0, 255, 0},
		{255, 255, 0},
		{0, 0, 255},
		{255, 0, 255},
		{0, 255, 255},
		{255, 255, 255},
	}
)

func init() {
	// Initialize colors from index 16 to 231.  These are a 6x6x6 cubee.
	for i := 16; i < 232; i++ {
		r := (i - 16) / 36
		g := ((i - 16) / 6) % 6
		b := (i - 16) % 6
		Palette[i].R = uint8((r * 255) / 5)
		Palette[i].G = uint8((g * 255) / 5)
		Palette[i].B = uint8((b * 255) / 5)
	}

	// Initialize colors from inidex 232 to 255.  These are grays in 24 steps.
	for i := 0; i < 24; i++ {
		g := uint8((i * 255) / 23)
		Palette[i+232] = RGB{g, g, g}
	}
}

// FindSimpleColor matches the r, g, b, color to one of the first 16 entries.
// The first 16 entries match the colors available in the 4-bit palette.
func FindSimpleColor(r, g, b uint8) int {
	color := 0
	dist := calculateDistance(RGB{r, g, b}, RGB{})

	for i := 1; i < 16; i++ {
		newDist := calculateDistance(RGB{r, g, b}, Palette[i])
		if newDist < dist {
			color = i
			dist = newDist
		}
	}

	return color
}

// FindPaletteColor matches the r, g, b, color to the closest color in the
// 8-bit palette.  The first 16 colors are not used, as the exact color for
// those entries are terminal dependent, and matching cannot be guarantted.
func FindPaletteColor(r, g, b uint8) int {
	color := 16
	dist := calculateDistance(RGB{r, g, b}, RGB{})

	// Avoid using the first 16 colors from the palette.  Terminals
	// disagree about the exact RGB for those colors, and the values in the
	// palette are just approximations.

	for i := 17; i < 0x100; i++ {
		newDist := calculateDistance(RGB{r, g, b}, Palette[i])
		if newDist < dist {
			color = i
			dist = newDist
		}
	}

	return color
}

func calculateDistance(a, b RGB) int32 {
	return sq(a.R, b.R) + sq(a.G, b.G) + sq(a.B, b.B)
}

func sq(a, b uint8) int32 {
	delta := int32(a) - int32(b)
	return delta * delta
}
