package plot

import (
	"fmt"
	"io"

	"git.sr.ht/~rj/sgr"
	"git.sr.ht/~rj/sgr/internal/pad"
)

// HorizontalBar returns an object which will show a horizontal bar when
// printed using a function in package fmt.  The parameter value must be
// in the range 0 to 1 inclusive, and sets the percentage of the bar that
// will be filled.
//
// To set the total width of the bar, use the width specifier in the format
// string.
func HorizontalBar(f *sgr.Formatter, value float64, color sgr.Color) interface{} {
	// Select options that will work when there is no ANSI support.
	if f == nil {
		return &hbar{
			value: value,
		}
	}

	// We can use full block (U+2588) to print, but that also imposes a
	// requirement for unicode support.  Unicode support is very likely,
	// but still an additional requirement.  Handle printing entirely
	// by manipulating the foreground and background colors.

	if color == sgr.Default || f.ColorDepth() <= 1 {
		return &hbar{
			value: value,
			style: f.NewStyle(sgr.Reverse),
		}
	}

	return &hbar{
		value: value,
		style: f.NewStyle(sgr.FG(color), sgr.BG(color)),
	}
}

type hbar struct {
	value float64
	style sgr.Style
}

func (hb *hbar) Format(f fmt.State, v rune) {
	// Get the width for the bar from the format state.
	// Apply a default if the user has not specified a width.
	width, ok := f.Width()
	if !ok {
		width = 10
	}

	// Determine how many characters to fill, and how many to pad.
	midpoint := mulRound(width, hb.value)

	// If the caller did not provide a value between 0 and 1, the midpoint
	// may not be a valid value.  Although the standard library often panics
	// on programming bugs, the package fmt takes a different approach.
	if midpoint < 0 || midpoint > width {
		fmt.Fprintf(f, "%%!(BADVALUE)%g", hb.value)
		return
	}

	// Note: All of the following write to the internal buffer that is part
	// of fmt.State.  No need to check for errors.
	if len(hb.style) > 0 {
		_, _ = f.Write(hb.style)
		pad.WriteSpaces(f, midpoint)
		_, _ = io.WriteString(f, "\x1b[0m" /* reset */)
		pad.WriteSpaces(f, width-midpoint)
	} else {
		const fillSolid = "################################"
		pad.WritePadding(f, midpoint, fillSolid)
		pad.WriteSpaces(f, width-midpoint)
	}
}
