package plot

import (
	"fmt"
	"io"

	"git.sr.ht/~rj/sgr"
	"git.sr.ht/~rj/sgr/internal/pad"
)

// ProgressBar returns an object which will show a progress bar when
// printed using a function in package fmt.  The parameter value must be
// in the range 0 to 1 inclusive, and sets the percentage of the bar that
// will be filled.
//
// To set the total width of the bar, use the width specifier in the format
// string.
func ProgressBar(f *sgr.Formatter, value float64, fg, bg sgr.Color) interface{} {
	// Select options that will work when there is no ANSI support.
	if f == nil {
		return &pbar{
			value: value,
		}
	}

	return &pbar{
		value: value,
		style: f.NewStyle(sgr.FG(fg), sgr.BG(bg)),
	}
}

type pbar struct {
	value float64
	style sgr.Style
}

func (pb *pbar) Format(f fmt.State, v rune) {
	// Get the width for the bar from the format state.
	// Apply a default if the user has not specified a width.
	width, ok := f.Width()
	if !ok {
		width = 10
	}

	// Determine how many characters to fill, and how many to pad.
	midpoint := mulRound(width, pb.value)

	// If the caller did not provide a value between 0 and 1, the midpoint
	// may not be a valid value.  Although the standard library often panics
	// on programming bugs, the package fmt takes a different approach.
	if midpoint < 0 || midpoint > width {
		fmt.Fprintf(f, "%%!(BADVALUE)%g", pb.value)
		return
	}

	// Note: All of the following write to the internal buffer that is part
	// of fmt.State.  No need to check for errors.
	if len(pb.style) > 0 {
		_, _ = f.Write(pb.style)
	}
	if midpoint > 1 {
		if haveUTF8 && len(pb.style) > 0 {
			pad.WritePadding(f, (midpoint-1)*len("≡"),
				"≡≡≡≡≡≡≡≡≡≡≡≡≡≡")
			_, _ = io.WriteString(f, "\u25B6")
		} else {
			pad.WritePadding(f, midpoint-1, "================")
			_, _ = io.WriteString(f, ">")
		}
	}
	pad.WriteSpaces(f, width-midpoint)
	if len(pb.style) > 0 {
		_, _ = io.WriteString(f, "\x1b[0m" /* reset */)
	}
}
