package sgr

import (
	"fmt"
	"io"

	"git.sr.ht/~rj/sgr/internal/pad"
)

var (
	sgrReset   = []byte("\x1b[0m")
	sgrBold    = []byte("\x1b[1m")
	sgrDim     = []byte("\x1b[2m")
	sgrRedFG   = []byte("\x1b[31m")
	sgrGreenFG = []byte("\x1b[32m")
)

// Bold returns an object which will show the text in bold when printed using
// a function in package fmt.
func (f *Formatter) Bold(text string) interface{} {
	if f == nil {
		return text
	}

	return boldFormatter(text)
}

type boldFormatter string

func (s boldFormatter) Format(f fmt.State, v rune) {
	format(
		f, v,
		sgrBold,
		string(s),
	)
}

// Dim returns an object which will show the text dimmed when printed using
// a function in package fmt.
func (f *Formatter) Dim(text string) interface{} {
	if f == nil {
		return text
	}

	return dimFormatter(text)
}

type dimFormatter string

func (s dimFormatter) Format(f fmt.State, v rune) {
	format(
		f, v,
		sgrDim,
		string(s),
	)
}

// Red returns an object which will show the text in red when printed using
// a function in package fmt.
func (f *Formatter) Red(text string) interface{} {
	if f == nil || f.colorDepth <= 1 {
		return text
	}

	return redFormatter(text)
}

type redFormatter string

func (s redFormatter) Format(f fmt.State, v rune) {
	format(
		f, v,
		sgrRedFG,
		string(s),
	)
}

// Green returns an object which will show the text in green when printed using
// a function in package fmt.
func (f *Formatter) Green(text string) interface{} {
	if f == nil || f.colorDepth <= 1 {
		return text
	}

	return greenFormatter(text)
}

type greenFormatter string

func (s greenFormatter) Format(f fmt.State, v rune) {
	format(
		f, v,
		sgrGreenFG,
		string(s),
	)
}

func format(f fmt.State, v rune, sgr []byte, s string) {
	// Check that the verb is supported.  Otherwise, print error message.
	if v != 's' && v != 'v' {
		// Match the error reporting from the fmt package.
		// Note, this function is not always called from sgr.formatter,
		// but that is an internal detail.  Since all invocations support
		// the same verbs, this should be sufficient.
		fmt.Fprintf(f, "%%!%c(sgr.formatter=%v)", v, s)

		return
	}

	if prec, ok := f.Precision(); ok && len(s) > prec {
		s = s[:prec]
	}

	// We are writing to a memory buffer.  The following calls to io.Write will
	// not return an error.
	if width, ok := f.Width(); ok && len(s) < width {
		if f.Flag('-') {
			_, _ = f.Write(sgr)
			_, _ = io.WriteString(f, s)
			pad.WriteSpaces(f, width-len(s))
			_, _ = f.Write(sgrReset)
		} else {
			_, _ = f.Write(sgr)
			pad.WriteSpaces(f, width-len(s))
			_, _ = io.WriteString(f, s)
			_, _ = f.Write(sgrReset)
		}
	} else {
		_, _ = f.Write(sgr)
		_, _ = io.WriteString(f, s)
		_, _ = f.Write(sgrReset)
	}
}
