package sgr

import "git.sr.ht/~rj/sgr/internal/palette"

// Style represents a set of display attributes to use when printing text.
// A style can combine attributes such as whether the text is bold or dimmed,
// as well as setting foreground and background colors.
//
// A Style created by a Formatter should not be used with different Formatters.
// The style encodes parameters that may be specific to the terminal's
// capabilities.
type Style []byte

// NewStyle creates a new style by combining the specified display attributes.
// Display attributes can alter formatting, like bold and italic, or specify
// colors.
//
// Not all terminals will support all attributes.  In particular, colors will
// be mapped to lower color-width values to match the terminal's support.
func (f *Formatter) NewStyle(ops ...StyleOption) Style {
	// If the formatter is nil, then don't create any escape sequences.
	if f == nil {
		return nil
	}

	// If no options are provided, return the escape sequence for a reset.
	if len(ops) == 0 {
		return sgrReset
	}

	if f.colorDepth <= 1 {
		ops = removeColorOptions(ops)
		if len(ops) == 0 {
			return nil
		}
	} else if f.colorDepth <= 4 {
		ops = filterColorOptionsTo4(ops)
	} else if f.colorDepth <= 8 {
		ops = filterColorOptionsTo8(ops)
	}

	sb := make([]byte, 0, 16)

	// Prefix for the escape sequence
	sb = append(sb, "\x1b["...)

	// Apply first option
	sb = writeCommands(sb, ops[0])

	// Apply rest of options
	for _, v := range ops[1:] {
		sb = append(sb, ';')
		sb = writeCommands(sb, v)
	}

	// Suffix to terminate the escape sequence
	sb = append(sb, 'm')

	return (Style)(sb)
}

func removeColorOptions(ops []StyleOption) []StyleOption {
	n := 0
	for ; n < len(ops) && !ops[n].isColor(); n++ {
	}
	if n >= len(ops) {
		return ops
	}
	for _, v := range ops[n+1:] {
		if !v.isColor() {
			ops[n] = v
			n++
		}
	}
	return ops[:n]
}

func filterColorOptionsTo4(ops []StyleOption) []StyleOption {
	for i, v := range ops {
		switch v & 0xFF {
		case 0x80:
			clr := v.palette()
			if clr > 15 {
				rgb := palette.Palette[clr]
				clr = palette.FindSimpleColor(rgb.R, rgb.G, rgb.B)
			}
			ops[i] = FG(adjustSimpleColor(clr))
		case 0x81:
			clr := v.palette()
			if clr > 15 {
				rgb := palette.Palette[clr]
				clr = palette.FindSimpleColor(rgb.R, rgb.G, rgb.B)
			}
			ops[i] = BG(adjustSimpleColor(clr))
		case 0x82:
			r, g, b := v.rgb()
			ops[i] = FG(adjustSimpleColor(palette.FindSimpleColor(r, g, b)))
		case 0x83:
			r, g, b := v.rgb()
			ops[i] = BG(adjustSimpleColor(palette.FindSimpleColor(r, g, b)))
		}
	}

	return ops
}

func adjustSimpleColor(clr int) Color {
	if clr >= 8 {
		clr = clr + int(Gray) - 8
	}
	return Color(clr)
}

func filterColorOptionsTo8(ops []StyleOption) []StyleOption {
	for i, v := range ops {
		switch v & 0xFF {
		case 0x82:
			r, g, b := v.rgb()
			ops[i] = FG8(PaletteColor(palette.FindPaletteColor(r, g, b)))
		case 0x83:
			r, g, b := v.rgb()
			ops[i] = BG8(PaletteColor(palette.FindPaletteColor(r, g, b)))
		}
	}

	return ops
}

func writeCommands(sb []byte, op StyleOption) []byte {
	// The standard escape codes are all 7-bit value.
	// We still need to adjust depending on the number of decimal digits.
	code := uint8(op)
	if code < 10 {
		return append(sb, '0'+code)
	}
	if code < 100 {
		return append(sb, '0'+(code/10), '0'+(code%10))
	}
	if code < 0x80 {
		return append(sb, '1', '0'+((code/10)%10), '0'+(code%10))
	}

	// The following codes are not in the ANSI standard, but are used by this
	// package to encode color commands that require multiple codes.
	if code == 0x80 {
		sb = append(sb, "38;5;"...)
		return writeCommand(sb, byte(op>>8))
	}
	if code == 0x81 {
		sb = append(sb, "48;5;"...)
		return writeCommand(sb, byte(op>>8))
	}
	if code == 0x82 {
		sb = append(sb, "38;2;"...)
		sb = writeCommand(sb, byte(op>>8))
		sb = append(sb, ';')
		sb = writeCommand(sb, byte(op>>16))
		sb = append(sb, ';')
		sb = writeCommand(sb, byte(op>>24))
		return sb
	}
	if code == 0x83 {
		sb = append(sb, "48;2;"...)
		sb = writeCommand(sb, byte(op>>8))
		sb = append(sb, ';')
		sb = writeCommand(sb, byte(op>>16))
		sb = append(sb, ';')
		sb = writeCommand(sb, byte(op>>24))
		return sb
	}

	return sb
}

func writeCommand(sb []byte, op byte) []byte {
	if op < 10 {
		return append(sb, '0'+op)
	}
	if op < 100 {
		return append(sb, '0'+(op/10), '0'+(op%10))
	}
	return append(sb, '0'+(op/100), '0'+((op/10)%10), '0'+(op%10))
}

// StyleOption encodes a single display attribute.
type StyleOption uint32

// Display attributes available for use with NewStyle.
const (
	Bold       StyleOption = 1 // Set the style to bold.
	Dim        StyleOption = 2 // Set the style to dim.
	Italic     StyleOption = 3 // Set the style to italic.
	Underline  StyleOption = 4 // Set the style to underline.
	Blink      StyleOption = 5 // Set the style to blink.
	Reverse    StyleOption = 7 // Set the style to reverse video.
	CrossedOut StyleOption = 9 // Set the style to crossed-out.
)

// FG creates a StyleOption that will set the foreground color to one of the
// colors in the 4-bit palette.
func FG(clr Color) StyleOption {
	return StyleOption(clr + 30)
}

// FG8 creates a StyleOption that will set the foreground color to one of the
// colors in the 8-bit palette.
func FG8(index PaletteColor) StyleOption {
	return 0x80 | (StyleOption(index) << 8)
}

// FG24 creates a StyleOption that will set the foreground color.
func FG24(r, g, b uint8) StyleOption {
	return 0x82 |
		(StyleOption(r) << 8) |
		(StyleOption(g) << 16) |
		(StyleOption(b) << 24)
}

// BG creates a StyleOption that will set the background color to one of the
// colors in the 4-bit palette.
func BG(clr Color) StyleOption {
	return StyleOption(clr + 40)
}

// BG8 creates a StyleOption that will set the background color to one of the
// colors in the 8-bit palette.
func BG8(index PaletteColor) StyleOption {
	return 0x81 | StyleOption(index)<<8
}

// BG24 creates a StyleOption that will set the background color.
func BG24(r, g, b uint8) StyleOption {
	return 0x83 |
		(StyleOption(r) << 8) |
		(StyleOption(g) << 16) |
		(StyleOption(b) << 24)
}

func (op StyleOption) isColor() bool {
	i := byte(op)
	return (i >= 30 && i <= 49) || (i >= 90 && i <= 107) || (i&0x80) != 0
}

func (op StyleOption) palette() int {
	return int((op >> 8) & 0xff)
}

func (op StyleOption) rgb() (r, g, b uint8) {
	r = uint8(op >> 8)
	g = uint8(op >> 16)
	b = uint8(op >> 24)
	return r, g, b
}
